@@ -1,7 +1,7 @@
 /**
  * @file ttrss_source.c  tt-rss feed list source support
  * 
- * Copyright (C) 2010-2011 Lars Lindner <lars.lindner@gmail.com>
+ * Copyright (C) 2010-2013 Lars Windolf <lars.lindner@gmail.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -34,25 +34,21 @@
 #include "node.h"
 #include "subscription.h"
 #include "update.h"
-#include "ui/auth_dialog.h"
 #include "ui/liferea_dialog.h"
-#include "ui/ui_common.h"
 #include "fl_sources/node_source.h"
 #include "fl_sources/opml_source.h"
 
-// FIXME: Avoid doing requests when we are not logged in yet!
-
-/** default tt-rss subscription list update interval = once a day */
-#define TTRSS_SOURCE_UPDATE_INTERVAL 60*60*24
-
-/** create a tt-rss source with given node as root */ 
+/** Initialize a TinyTinyRSS source with given node as root */ 
 static ttrssSourcePtr
 ttrss_source_new (nodePtr node) 
 {
 	ttrssSourcePtr source = g_new0 (struct ttrssSource, 1) ;
-	source->root = node; 
+	source->root = node;
+	source->apiLevel = 0;
 	source->actionQueue = g_queue_new (); 
-	source->loginState = TTRSS_SOURCE_STATE_NONE; 
+	source->loginState = TTRSS_SOURCE_STATE_NONE;
+	source->categories = g_hash_table_new (g_direct_hash, g_direct_equal);
+	source->categoryNodes = g_hash_table_new (g_direct_hash, g_direct_equal);
 	
 	return source;
 }
@@ -71,41 +67,18 @@ ttrss_source_free (ttrssSourcePtr source)
 }
 
 static void
-ttrss_source_get_config_cb (const struct updateResult * const result, gpointer userdata, updateFlags flags)
+ttrss_source_set_login_error (ttrssSourcePtr source, gchar *msg)
 {
-	ttrssSourcePtr	source = (ttrssSourcePtr) userdata;
-	subscriptionPtr subscription = source->root->subscription;
+	g_free (source->root->subscription->updateError);
+	source->root->subscription->updateError = msg;
+	source->root->available = FALSE;
 
-	debug1 (DEBUG_UPDATE, "tt-rss getConfig processing... >>>%s<<<", result->data);
+	g_warning ("TinyTinyRSS login failed: error '%s'!\n", msg);
 
-	/* We expect something like:
-
-		 {"icons_dir":"icons","icons_url":"icons","daemon_is_running":true,"num_feeds":71}
-
-	   And are only interested in the "daemon_is_running" value... */
-	
-	if (result->data && result->httpstatus == 200) {
-		JsonParser *parser = json_parser_new ();
-
-		if (json_parser_load_from_data (parser, result->data, -1, NULL)) {
-			JsonNode *node = json_parser_get_root (parser);
-			const gchar *result = json_get_string (node, "daemon_is_running");
-			if (result && g_str_equal ("true", result)) {
-				source->selfUpdating = TRUE;
-				debug0 (DEBUG_UPDATE, "This tt-rss source is self-updating!");
-			} else {
-				debug0 (DEBUG_UPDATE, "This tt-rss source is not self-updating!");
-			}
-
-			g_object_unref (parser);
-		}
-	}
-
-	/* now that we are authenticated and know the config trigger updating to start data retrieval */
-	source->loginState = TTRSS_SOURCE_STATE_ACTIVE;
-
-	if (!(flags & TTRSS_SOURCE_UPDATE_ONLY_LOGIN) && !source->selfUpdating)
-		subscription_update (subscription, flags);
+	if (++source->authFailures < TTRSS_SOURCE_MAX_AUTH_FAILURES)
+		source->loginState = TTRSS_SOURCE_STATE_NONE;
+	else
+		source->loginState = TTRSS_SOURCE_STATE_NO_AUTH;
 }
 
 static void
@@ -113,48 +86,51 @@ ttrss_source_login_cb (const struct updateResult * const result, gpointer userda
 {
 	ttrssSourcePtr	source = (ttrssSourcePtr) userdata;
 	subscriptionPtr subscription = source->root->subscription;
-		
-	debug1 (DEBUG_UPDATE, "tt-rss login processing... >>>%s<<<", result->data);
+	JsonParser	*parser;
+	
+	debug1 (DEBUG_UPDATE, "TinyTinyRSS login processing... >>>%s<<<", result->data);
 	
 	g_assert (!source->session_id);
 	
-	if (result->data && result->httpstatus == 200) {
-		JsonParser *parser = json_parser_new ();
-		
-		if (json_parser_load_from_data (parser, result->data, -1, NULL)) {
-			JsonNode *node = json_parser_get_root (parser);
-
-			if (json_get_string (node, "error"))
-				g_warning ("tt-rss login failed: error '%s'!\n", json_get_string (node, "error"));
-			
-			source->session_id = g_strdup (json_get_string (json_get_node (node, "content"), "session_id"));
-			if (source->session_id) {
-				debug1 (DEBUG_UPDATE, "Found session_id: >>>%s<<<!", source->session_id);
-			} else {
-				g_warning ("No tt-rss session_id found in response!\n%s", result->data);
-			}
-			
-			g_object_unref (parser);
-		} else {
-			g_warning ("Invalid JSON returned on tt-rss login! >>>%s<<<", result->data);
-		}
-	} else {
-		g_warning ("tt-rss login failed: HTTP %d\n", result->httpstatus);
-		source->root->available = FALSE;
+	if (!(result->data && result->httpstatus == 200)) {
+		ttrss_source_set_login_error (source, g_strdup_printf ("Login request failed with HTTP error %d", result->httpstatus));
+		return;
 	}
 
+	parser = json_parser_new ();
+	if (!json_parser_load_from_data (parser, result->data, -1, NULL)) {
+		ttrss_source_set_login_error (source, g_strdup ("Invalid JSON returned on login!"));
+		return;
+	}
+	
+	JsonNode *node = json_parser_get_root (parser);
+
+	/* Check for API specified error... */
+	if (json_get_string (json_get_node (node, "content"), "error")) {
+		const gchar *error = json_get_string (json_get_node (node, "content"), "error");
+
+		ttrss_source_set_login_error (source, g_strdup (error));
+		if (g_str_equal (error, "LOGIN_ERROR"))
+			auth_dialog_new (source->root->subscription, flags);
+
+		return;
+	}
+
+	/* Success! Get SID and API Level. */
+	source->apiLevel = json_get_int (json_get_node (node, "content"), "api_level");
+	source->session_id = g_strdup (json_get_string (json_get_node (node, "content"), "session_id"));
 	if (source->session_id) {
-		updateRequestPtr request;
+		debug2 (DEBUG_UPDATE, "TinyTinyRSS Found session_id: >>>%s<<< (API level %d)!", source->session_id, source->apiLevel);
 
-		/* Check for remote update daemon running. This needs to be known
-		   before we start updating to decide wether to actively update
-		   remote feeds or just fetch them. */
-		request = update_request_new ();
-		request->options = update_options_copy (subscription->updateOptions);
+		source->loginState = TTRSS_SOURCE_STATE_ACTIVE;
 
-		update_request_set_source (request, g_strdup_printf (TTRSS_GET_CONFIG, metadata_list_get (subscription->metadata, "ttrss-url"), source->session_id));
-		update_execute_request (source, request, ttrss_source_get_config_cb, source, flags);
+		if (!(flags & TTRSS_SOURCE_UPDATE_ONLY_LOGIN))
+			subscription_update (subscription, flags);
+	} else {
+		ttrss_source_set_login_error (source, g_strdup_printf ("No session_id found in response!\n%s", result->data));
 	}
+	
+	g_object_unref (parser);
 }
 
 /**
@@ -164,25 +140,34 @@ ttrss_source_login_cb (const struct updateResult * const result, gpointer userda
 void
 ttrss_source_login (ttrssSourcePtr source, guint32 flags) 
 { 
-	gchar			*username, *password;
+	gchar			*username, *password, *source_uri;
 	updateRequestPtr	request;
 	subscriptionPtr		subscription = source->root->subscription;
 	
 	if (source->loginState != TTRSS_SOURCE_STATE_NONE) {
 		/* this should not happen, as of now, we assume the session doesn't expire. */
-		debug1(DEBUG_UPDATE, "Logging in while login state is %d", source->loginState);
+		debug1 (DEBUG_UPDATE, "Logging in while login state is %d", source->loginState);
 	}
 
 	request = update_request_new ();
 
-	/* escape user and password as both are passed using an URI */
-	username = g_uri_escape_string (subscription->updateOptions->username, NULL, TRUE);
-	password = g_uri_escape_string (subscription->updateOptions->password, NULL, TRUE);
-	
-	update_request_set_source (request, g_strdup_printf (TTRSS_LOGIN_URL, metadata_list_get (subscription->metadata, "ttrss-url"), username, password));
+	/* escape user and password for JSON call */
+	username = g_strescape (subscription->updateOptions->username, NULL);
+	password = g_strescape (subscription->updateOptions->password, NULL);
+
+	source->url = metadata_list_get (subscription->metadata, "ttrss-url");
+	if (!source->url) {
+		ttrss_source_set_login_error (source, g_strdup ("Fatal: We've lost the TinyTinyRSS server URL! Please re-subscribe!"));
+		return;
+	}
+
+	source_uri = g_strdup_printf (TTRSS_URL, source->url);
+	update_request_set_source (request, source_uri);
+	g_free (source_uri);
 
 	request->options = update_options_copy (subscription->updateOptions);
-	
+	request->postdata = g_strdup_printf (TTRSS_JSON_LOGIN, username, password);
+
 	g_free (username);
 	g_free (password);
 
@@ -196,22 +181,20 @@ ttrss_source_login (ttrssSourcePtr source, guint32 flags)
 static void
 ttrss_source_update (nodePtr node)
 {
-	debug0(DEBUG_UPDATE, "ttrss_source_update()");
+	debug0 (DEBUG_UPDATE, "ttrss_source_update()");
 	subscription_update (node->subscription, 0);
 }
 
 static void
 ttrss_source_auto_update (nodePtr node)
 {
-	GTimeVal	now;
 	ttrssSourcePtr	source = (ttrssSourcePtr) node->data;
 
 	if (source->loginState == TTRSS_SOURCE_STATE_IN_PROGRESS) 
 		return; /* the update will start automatically anyway */
 
-	g_get_current_time (&now);
-
-	ttrss_source_update (node);
+	debug0 (DEBUG_UPDATE, "ttrss_source_auto_update()");
+	subscription_auto_update (node->subscription);
 }
 
 static void
@@ -223,18 +206,31 @@ ttrss_source_init (void)
 
 static void ttrss_source_deinit (void) { }
 
+static void
+ttrss_source_set_subscription_type (nodePtr folder)
+{
+	GSList *iter;
+
+	for (iter = folder->children; iter; iter = g_slist_next(iter)) {
+		nodePtr node = (nodePtr) iter->data;
+
+		if (node->subscription)
+			node->subscription->type = &ttrssSourceFeedSubscriptionType;
+		else
+			ttrss_source_set_subscription_type (node);
+	}
+}
+
 static void
 ttrss_source_import (nodePtr node)
 {
-	GSList *iter; 
 	opml_source_import (node);
-	
+
 	node->subscription->type = &ttrssSourceSubscriptionType;
 	if (!node->data)
 		node->data = (gpointer) ttrss_source_new (node);
 
-	for (iter = node->children; iter; iter = g_slist_next(iter))
-		((nodePtr) iter->data)->subscription->type = &ttrssSourceFeedSubscriptionType;
+	ttrss_source_set_subscription_type (node);
 }
 
 static void
@@ -266,23 +262,26 @@ on_ttrss_source_selected (GtkDialog *dialog,
 		nodePtr		node;
 		subscriptionPtr subscription = subscription_new ("", NULL, NULL);
 		
-		/* The is a bit ugly: we need to prevent the tt-rss base
+		/* This is a bit ugly: we need to prevent the tt-rss base
 		   URL from being lost by unwanted permanent redirects on
 		   the getFeeds call, so we save it as the homepage meta
 		   data value... */
 		metadata_list_set (&subscription->metadata, "ttrss-url", gtk_entry_get_text (GTK_ENTRY (liferea_dialog_lookup (GTK_WIDGET (dialog), "serverUrlEntry"))));
-		
-		subscription->updateOptions->username = g_strdup (gtk_entry_get_text (GTK_ENTRY (liferea_dialog_lookup (GTK_WIDGET(dialog), "userEntry"))));
-		subscription->updateOptions->password = g_strdup (gtk_entry_get_text (GTK_ENTRY (liferea_dialog_lookup (GTK_WIDGET(dialog), "passwordEntry"))));
-		subscription->type = &ttrssSourceSubscriptionType;
-		
+
 		node = node_new (node_source_get_node_type ());
 		node_set_title (node, "Tiny Tiny RSS");
 		node_source_new (node, ttrss_source_get_type ());
 		node_set_subscription (node, subscription);
-		node->data = ttrss_source_new (node);
+		
+		subscription_set_auth_info (subscription,
+		                            gtk_entry_get_text (GTK_ENTRY (liferea_dialog_lookup (GTK_WIDGET(dialog), "userEntry"))),
+		                            gtk_entry_get_text (GTK_ENTRY (liferea_dialog_lookup (GTK_WIDGET(dialog), "passwordEntry"))));
+		subscription->type = &ttrssSourceSubscriptionType;		
+
+		node->data = (gpointer)ttrss_source_new (node);
 		feedlist_node_added (node);
 		ttrss_source_update (node);
+		db_node_update (node);
 	}
 
 	gtk_widget_destroy (GTK_WIDGET (dialog));
@@ -305,6 +304,8 @@ static void
 ttrss_source_cleanup (nodePtr node)
 {
 	ttrssSourcePtr source = (ttrssSourcePtr) node->data;
+	g_hash_table_destroy (source->categories);
+	g_hash_table_destroy (source->categoryNodes);
 	ttrss_source_free (source);
 	node->data = NULL;
 }
@@ -312,7 +313,7 @@ ttrss_source_cleanup (nodePtr node)
 static void
 ttrss_source_remote_update_cb (const struct updateResult * const result, gpointer userdata, updateFlags flags)
 {
-	debug2 (DEBUG_UPDATE, "tt-rss result processing... status:%d >>>%s<<<", result->httpstatus, result->data);
+	debug2 (DEBUG_UPDATE, "TinyTinyRSS result processing... status:%d >>>%s<<<", result->httpstatus, result->data);
 }
 
 /* FIXME: Only simple synchronous item change requests... Get async! */
@@ -326,12 +327,9 @@ ttrss_source_item_set_flag (nodePtr node, itemPtr item, gboolean newStatus)
 
 	request = update_request_new ();
 	request->options = update_options_copy (root->subscription->updateOptions);
+	request->postdata = g_strdup_printf (TTRSS_JSON_UPDATE_ITEM_FLAG, source->session_id, item_get_id(item), newStatus?1:0 );
 
-	update_request_set_source (request, g_strdup_printf (TTRSS_UPDATE_ITEM_FLAG,
-	                                       metadata_list_get (root->subscription->metadata, "ttrss-url"),
-	                                       source->session_id, 
-	                                       item_get_id (item), newStatus?1:0));
-
+	update_request_set_source (request, g_strdup_printf (TTRSS_URL, source->url));
 	update_execute_request (source, request, ttrss_source_remote_update_cb, source, 0 /* flags */);
 
 	item_flag_state_changed (item, newStatus);
@@ -346,12 +344,9 @@ ttrss_source_item_mark_read (nodePtr node, itemPtr item, gboolean newStatus)
 
 	request = update_request_new ();
 	request->options = update_options_copy (root->subscription->updateOptions);
+	request->postdata = g_strdup_printf (TTRSS_JSON_UPDATE_ITEM_UNREAD, source->session_id, item_get_id(item), newStatus?0:1 );
 
-	update_request_set_source (request, g_strdup_printf (TTRSS_UPDATE_ITEM_UNREAD,
-	                                       metadata_list_get (root->subscription->metadata, "ttrss-url"),
-	                                       source->session_id, 
-	                                       item_get_id (item), newStatus?0:1));
-
+	update_request_set_source (request, g_strdup_printf (TTRSS_URL, source->url));
 	update_execute_request (source, request, ttrss_source_remote_update_cb, source, 0 /* flags */);
 
 	item_read_state_changed (item, newStatus);
@@ -364,7 +359,7 @@ static struct nodeSourceType nst = {
 	.name                = N_("Tiny Tiny RSS"),
 	.description         = N_("Integrate the feed list of your Tiny Tiny RSS 1.5+ account. Liferea will "
 	   "present your tt-rss subscriptions, and will synchronize your feed list and reading lists."),
-	.capabilities        = NODE_SOURCE_CAPABILITY_DYNAMIC_CREATION | 
+	.capabilities        = NODE_SOURCE_CAPABILITY_DYNAMIC_CREATION |
 	                       NODE_SOURCE_CAPABILITY_ITEM_STATE_SYNC,
 	.source_type_init    = ttrss_source_init,
 	.source_type_deinit  = ttrss_source_deinit,
@@ -380,7 +375,8 @@ static struct nodeSourceType nst = {
 	.item_mark_read      = ttrss_source_item_mark_read,
 	.add_folder          = NULL,	/* not supported by current tt-rss JSON API (v1.5) */
 	.add_subscription    = NULL,	/* not supported by current tt-rss JSON API (v1.5) */
-	.remove_node         = NULL	/* not supported by current tt-rss JSON API (v1.5) */
+	.remove_node         = NULL,	/* not supported by current tt-rss JSON API (v1.5) */
+	.convert_to_local    = NULL	/* FIXME: implement me to allow data migration from tt-rss! */
 };
 
 nodeSourceTypePtr
